#!/usr/bin/env python

import sys, os, string, getopt, stat

# This is the number of accesses needed to ignore a file
MAX_ACCESSES = 10

def help(retval = 1):
    print '''
Crystal Space CVS Update script, Version 1.0
Written by Michael Dale Long.
Copyright(c) 1999 under the terms of the GNU Lesser General
Public License.

Usage:  bin/csupdate.py [OPTIONS]
Options:
-b, --build         Execute bin/build.pl after update is completed
-h, --help          Display this help screen
-d, --detailed-help Display a more detailed help screen
-v, --version       Display the version number
-V, --verbose       Print CVS error messages

NOTE:  This should only be run from the root CS directory.  Also,
the --build and --verbose options may make use of UNIX specific
features, let me know if you have problems using them (and better
yet, fixes for them! :)
'''
    sys.exit(retval)

def detailed_help(retval = 1):
    print '''
Crystal Space CVS Update script, Version 1.0
Written by Michael Dale Long.
Copyright(c) 1999 under the terms of the GNU Lesser General
Public License.

I wrote this script to ease the CVS update/build cycle, and also to create a
way to cut out the extra junk that CVS always spits out when it updates.  This
script 'memorizes' what files you never change, and after so many updates it 
doesn't include them in the output or the log file it generates.  This allows
you to see what files are in the source tree that are significant, and letting
know when something that doesn't normally change has changed.

Relevant files:
bin/csupdate.py - The script itself
bin/cvslog.py - The script's memory file
cvs.log - The log file of CVS transactions

That last two are automatically generated and overwritten everytime you update
from CVS with the script.

Usage:  bin/csupdate.py [OPTIONS]"
Options:
-b, --build         Execute bin/build.pl after update is completed
-h, --help          Display a terser help screen
-d, --detailed-help Displays this help screen
-v, --version       Display the version number
-V, --verbose       Print CVS error messages

bin/csupdate.py expects to be run from the root CS directory.  Also,
the --build and --verbose options may make use of UNIX specific
features.  Let me know if you have troubles/successes on other 
platforms.  Patches are welcome, too.

The --version, --help, and --detailed-help options are obvious information
options.  The script doesn't do anything else if they are specified.

The --verbose option tell CVS to print out error messages.  Specifically, it
doesn't redirect stderr to /dev/null (that's the UNIX dependent part).

The --build option executes another script to build CS.  Very nice for 
updating and compiling in one fell swoop, although for really automated 
builds be sure to change the line
    $Make="make all";
to
    $Make="make -k all";
so errors don't stop the build process.  Check the other options there to see
if they're to your liking as well.  A text build.log and 
platform-processor.html (e.g. linux-i386.html) file are created to log the 
build process, for your convenience.  Oh yeah, it's a Perl script, so you'll
need Perl installed to make it work.

One thing to note, since csupdate always creates new versions of cvs.log and
bin/cvslog.py each time it's run, it's best to add those to your .cvsignore
file for the root and bin directories, respectively.

Finally, here are some bash aliases I use to make my CS life easier:

alias update-cs="cd <wherever CS is> ; ./bin/csupdate.py"
alias build-cs="cd <wherever CS is> ; ./bin/build.pl"
alias update-build-cs="update-cs --build"

Todo:

* Add a --silence option?
* Add a --no-log option?
These might come in handy in some case, I don't know.
* If a changed file is encountered ("C", conflicting merge),
  disable --build and warn the user
'''
    sys.exit(retval)

def do_ignore(cvsfile):
    if cvsdict.has_key(cvsfile):
	# Decode the dictionary tuple
	xs, date = cvsdict[cvsfile]
	# Get the modification time for the file
	cvstat = os.stat(cvsfile[2:])
	# See if the file has been modified since the date in memory
	if(date > cvstat[stat.ST_MTIME]):
	    # Encode the new time in memory
	    cvsnew[cvsfile] = 1, cvstat[stat.ST_MTIME]
	    return 0 # Don't ignore
	elif(xs > MAX_ACCESSES):
	    # Save the old entry
	    cvsnew[cvsfile] = cvsdict[cvsfile]
	    return 1 # Ignore
	else:
	    # Save the old date and bump up the access number
	    cvsnew[cvsfile] = xs + 1, date
	    return 0 # Don't ignore
    else:
	try:
	    # Get the modification time
	    cvstat = os.stat(cvsfile[2:])
	    # Encode the new entry
	    cvsnew[cvsfile] = 1, cvstat[stat.ST_MTIME]
	except OSError:
	    # This message probably wasn't a file at all
	    print "WARNING:  Unknown message from CS:"
	    print cvsfile
	    print
	return 0 # Don't ignore

# Process the argument and option list
try:
    optlist, args = getopt.getopt(sys.argv[1:], "Vbhdv", ["verbose", "build", "help", "detailed-help", "version"])
except getopt.error:
    print "ERROR:  Unrecognized option!"
    print
    help()

# There should be no arguments
if len(args)!=0:
    print "ERROR:  Too many arguments!"
    print
    help()

# State variables
build = 0                       # Corresponds to --build
verbose = " 2> /dev/null"       # Corresponds to --verbose

# Process options
for i in range(len(optlist)):
    if optlist[i][0] == "--build" or optlist[i][0] == "-b":
	build = 1
    if optlist[i][0] == "--help" or optlist[i][0] == "-h":
	help(0)
    if optlist[i][0] == "--detailed-help" or optlist[i][0] == "-d":
	detailed_help(0)
    if optlist[i][0] == "--version" or optlist[i][0] == "-v":
	print "Crystal Space CVS Update script, Version 1.0"
	sys.exit()
    if optlist[i][0] == "--verbose" or optlist[i][0] == "-V":
	verbose = ""

# Load up the optional bin/cvslog.py.  It is the "memory" for csupdate.py.
# It keeps track of what files to ignore when writing the display and logfile.
try:
    execfile("bin/cvslog.py")
except IOError:
    cvsmem = {}

# Initialize the new memory list
cvsnew = {}

# Open the cvs.log file.
try:
    cvslog = open("cvs.log", "w")
except IOError:
    print "ERROR:  Can't open cvs.log for overwrite!"
    sys.exit(1)

# Execute cvs update.  verbose specifies what to do with stderr, see the options above.
try:
    cvsin = os.popen("cvs -z9 update" + verbose)
except IOError:
    print "ERROR:  Can't execute cvs!"
    sys.exit(1)

# Process cvs output
curline = cvsin.readline()
while curline!="":
    if do_ignore(curline[:-1])==0:
	print curline[:-1]
	cvslog.write(curline)
    curline = cvsin.readline()

# Finish up output
print curline[:-1]
cvslog.write(curline)

# Close all files, and report if CVS complained about anything
cvsresult = cvsin.close()
if cvsresult!=None:
    print "WARNING:  cvs exitted with abnormal code " + str(cvsresult)
cvslog.close

# Write out the new memory
try:
    cvsignore = open("bin/cvslog.py", "w+")
    cvsignore.write("cvsdict = " + repr(cvsnew))
    cvsignore.close()
except IOError:
    print "ERROR:  Couldn't overwrite cvslog.py!  Changes from this session lost!"

# Build only if CVS was successful (and --build was specified)
if build and cvsresult==None:
    os.system("bin/build.pl")
