/*
    Copyright (C) 2001 by Jorrit Tyberghein
    Copyright (C) 2001 by W.C.A. Wijngaards

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "cssysdef.h"
#include "cssys/sysfunc.h"
#include "csgeom/math3d.h"
#include "csutil/parser.h"
#include "csutil/scanstr.h"
#include "spr2dldr.h"
#include "imesh/object.h"
#include "iengine/mesh.h"
#include "iengine/engine.h"
#include "iutil/plugin.h"
#include "imesh/sprite2d.h"
#include "ivideo/graph3d.h"
#include "qint.h"
#include "iutil/strvec.h"
#include "csutil/util.h"
#include "iutil/object.h"
#include "iengine/material.h"
#include "ivaria/reporter.h"
#include "iutil/objreg.h"
#include "iutil/eventh.h"
#include "iutil/comp.h"

CS_IMPLEMENT_PLUGIN

CS_TOKEN_DEF_START
  CS_TOKEN_DEF (ADD)
  CS_TOKEN_DEF (ALPHA)
  CS_TOKEN_DEF (COPY)
  CS_TOKEN_DEF (KEYCOLOR)
  CS_TOKEN_DEF (TILING)
  CS_TOKEN_DEF (MULTIPLY2)
  CS_TOKEN_DEF (MULTIPLY)
  CS_TOKEN_DEF (TRANSPARENT)
  CS_TOKEN_DEF (UVANIMATION)
  CS_TOKEN_DEF (FRAME)

  CS_TOKEN_DEF (COLORS)
  CS_TOKEN_DEF (FACTORY)
  CS_TOKEN_DEF (LIGHTING)
  CS_TOKEN_DEF (MATERIAL)
  CS_TOKEN_DEF (MIXMODE)
  CS_TOKEN_DEF (UV)
  CS_TOKEN_DEF (VERTICES)
  CS_TOKEN_DEF (ANIMATE)
CS_TOKEN_DEF_END

SCF_IMPLEMENT_IBASE (csSprite2DFactoryLoader)
  SCF_IMPLEMENTS_INTERFACE (iLoaderPlugin)
  SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iComponent)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE (csSprite2DFactoryLoader::eiComponent)
  SCF_IMPLEMENTS_INTERFACE (iComponent)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

SCF_IMPLEMENT_IBASE (csSprite2DFactorySaver)
  SCF_IMPLEMENTS_INTERFACE (iSaverPlugin)
  SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iComponent)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE (csSprite2DFactorySaver::eiComponent)
  SCF_IMPLEMENTS_INTERFACE (iComponent)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

SCF_IMPLEMENT_IBASE (csSprite2DLoader)
  SCF_IMPLEMENTS_INTERFACE (iLoaderPlugin)
  SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iComponent)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE (csSprite2DLoader::eiComponent)
  SCF_IMPLEMENTS_INTERFACE (iComponent)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

SCF_IMPLEMENT_IBASE (csSprite2DSaver)
  SCF_IMPLEMENTS_INTERFACE (iSaverPlugin)
  SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iComponent)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE (csSprite2DSaver::eiComponent)
  SCF_IMPLEMENTS_INTERFACE (iComponent)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

SCF_IMPLEMENT_FACTORY (csSprite2DFactoryLoader)
SCF_IMPLEMENT_FACTORY (csSprite2DFactorySaver)
SCF_IMPLEMENT_FACTORY (csSprite2DLoader)
SCF_IMPLEMENT_FACTORY (csSprite2DSaver)

SCF_EXPORT_CLASS_TABLE (spr2dldr)
  SCF_EXPORT_CLASS (csSprite2DFactoryLoader,
  	"crystalspace.mesh.loader.factory.sprite.2d",
	"Crystal Space Sprite2D Mesh Factory Loader")
  SCF_EXPORT_CLASS (csSprite2DFactorySaver,
  	"crystalspace.mesh.saver.factory.sprite.2d",
	"Crystal Space Sprite2D Mesh Factory Saver")
  SCF_EXPORT_CLASS (csSprite2DLoader, "crystalspace.mesh.loader.sprite.2d",
    "Crystal Space Sprite2D Mesh Loader")
  SCF_EXPORT_CLASS (csSprite2DSaver, "crystalspace.mesh.saver.sprite.2d",
    "Crystal Space Sprite2D Mesh Saver")
SCF_EXPORT_CLASS_TABLE_END

static void ReportError (iReporter* reporter, const char* id,
	const char* description, ...)
{
  va_list arg;
  va_start (arg, description);

  if (reporter)
  {
    reporter->ReportV (CS_REPORTER_SEVERITY_ERROR, id, description, arg);
  }
  else
  {
    char buf[1024];
    vsprintf (buf, description, arg);
    csPrintf ("Error ID: %s\n", id);
    csPrintf ("Description: %s\n", buf);
  }
  va_end (arg);
}

csSprite2DFactoryLoader::csSprite2DFactoryLoader (iBase* pParent)
{
  SCF_CONSTRUCT_IBASE (pParent);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scfiComponent);
  reporter = NULL;
  plugin_mgr = NULL;
}

csSprite2DFactoryLoader::~csSprite2DFactoryLoader ()
{
  if (reporter) reporter->DecRef ();
  SCF_DEC_REF (plugin_mgr);
}

bool csSprite2DFactoryLoader::Initialize (iObjectRegistry* object_reg)
{
  csSprite2DFactoryLoader::object_reg = object_reg;
  plugin_mgr = CS_QUERY_REGISTRY (object_reg, iPluginManager);
  reporter = CS_QUERY_REGISTRY (object_reg, iReporter);
  return true;
}

static UInt ParseMixmode (iReporter* reporter, char* buf)
{
  CS_TOKEN_TABLE_START (modes)
    CS_TOKEN_TABLE (COPY)
    CS_TOKEN_TABLE (MULTIPLY2)
    CS_TOKEN_TABLE (MULTIPLY)
    CS_TOKEN_TABLE (ADD)
    CS_TOKEN_TABLE (ALPHA)
    CS_TOKEN_TABLE (TRANSPARENT)
    CS_TOKEN_TABLE (KEYCOLOR)
    CS_TOKEN_TABLE (TILING)
  CS_TOKEN_TABLE_END

  char* name;
  long cmd;
  char* params;

  UInt Mixmode = 0;

  while ((cmd = csGetObject (&buf, modes, &name, &params)) > 0)
  {
    if (!params)
    {
      ReportError (reporter,
		"crystalspace.sprite2dloader.parse.mixmode.badformat",
		"Bad format while parsing mixmode!");
      return ~0;
    }
    switch (cmd)
    {
      case CS_TOKEN_COPY: Mixmode |= CS_FX_COPY; break;
      case CS_TOKEN_MULTIPLY: Mixmode |= CS_FX_MULTIPLY; break;
      case CS_TOKEN_MULTIPLY2: Mixmode |= CS_FX_MULTIPLY2; break;
      case CS_TOKEN_ADD: Mixmode |= CS_FX_ADD; break;
      case CS_TOKEN_ALPHA:
	Mixmode &= ~CS_FX_MASK_ALPHA;
	float alpha;
        csScanStr (params, "%f", &alpha);
	Mixmode |= CS_FX_SETALPHA(alpha);
	break;
      case CS_TOKEN_TRANSPARENT: Mixmode |= CS_FX_TRANSPARENT; break;
      case CS_TOKEN_KEYCOLOR: Mixmode |= CS_FX_KEYCOLOR; break;
      case CS_TOKEN_TILING: Mixmode |= CS_FX_TILING; break;
    }
  }
  if (cmd == CS_PARSERR_TOKENNOTFOUND)
  {
    ReportError (reporter,
		"crystalspace.sprite2dloader.parse.mixmode.badtoken",
		"Token '%s' not found while parsing mixmodes!",
		csGetLastOffender ());
    return ~0;
  }
  return Mixmode;
}

static void ParseAnim (iReporter* reporter, iSprite2DFactoryState* spr2dLook,
	const char *animname, char *buf)
{
  CS_TOKEN_TABLE_START (commands)
    CS_TOKEN_TABLE (FRAME)
  CS_TOKEN_TABLE_END

  char* name;
  long cmd;
  char* params;
  float verts[200];
  int duration;

  iSprite2DUVAnimation *ani = spr2dLook->CreateUVAnimation ();
  ani->SetName (animname);

  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    if (!params)
    {
      ReportError (reporter,
		"crystalspace.sprite2dfactoryloader.parse.badtoken",
		"Expected FRAME instead of '%s'!", buf);
      return;
    }
    switch (cmd)
    {
      case CS_TOKEN_FRAME:
	{
	  int num;
          csScanStr (params, "%d, %F", &duration, &verts, &num);
	  iSprite2DUVAnimationFrame *frame = ani->CreateFrame (-1);
	  frame->SetFrameData (name, duration, num/2, verts);
	}
	break;
    }
  }
  if (cmd == CS_PARSERR_TOKENNOTFOUND)
  {
    ReportError (reporter,
		"crystalspace.sprite2dfactoryloader.parse.badtoken",
		"Token '%s' not found while parsing FRAME!",
		csGetLastOffender ());
    return;
  }
}

iBase* csSprite2DFactoryLoader::Parse (const char* string, iEngine* engine,
	iBase* /* context */)
{
  CS_TOKEN_TABLE_START (commands)
    CS_TOKEN_TABLE (LIGHTING)
    CS_TOKEN_TABLE (MATERIAL)
    CS_TOKEN_TABLE (MIXMODE)
    CS_TOKEN_TABLE (UVANIMATION)
  CS_TOKEN_TABLE_END

  char* name;
  long cmd;
  char* params;
  char str[255];

  iMeshObjectType* type = CS_QUERY_PLUGIN_CLASS (plugin_mgr,
  	"crystalspace.mesh.object.sprite.2d", iMeshObjectType);
  if (!type)
  {
    type = CS_LOAD_PLUGIN (plugin_mgr, "crystalspace.mesh.object.sprite.2d",
    	iMeshObjectType);
  }
  if (!type)
  {
    ReportError (reporter,
		"crystalspace.sprite2dfactoryloader.setup.objecttype",
		"Could not load the sprite.2d mesh object plugin!");
    return NULL;
  }
  iMeshObjectFactory* fact = type->NewFactory ();
  type->DecRef ();
  iSprite2DFactoryState* spr2dLook = SCF_QUERY_INTERFACE (fact,
  	iSprite2DFactoryState);

  char* buf = (char*)string;
  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    if (!params)
    {
      ReportError (reporter,
		"crystalspace.sprite2dfactoryloader.parse.badformat",
		"Bad format while parsing sprite2d factory!");
      fact->DecRef ();
      spr2dLook->DecRef ();
      return NULL;
    }
    switch (cmd)
    {
      case CS_TOKEN_MATERIAL:
	{
          csScanStr (params, "%s", str);
          iMaterialWrapper* mat = engine->GetMaterialList ()->
	  	FindByName (str);
	  if (!mat)
	  {
	    ReportError (reporter,
		"crystalspace.sprite2dfactoryloader.parse.unknownmaterial",
		"Couldn't find material named '%s'", str);
            fact->DecRef ();
	    spr2dLook->DecRef ();
            return NULL;
	  }
	  spr2dLook->SetMaterialWrapper (mat);
	}
	break;
      case CS_TOKEN_LIGHTING:
        {
          bool do_lighting;
          csScanStr (params, "%b", &do_lighting);
          spr2dLook->SetLighting (do_lighting);
        }
	break;
      case CS_TOKEN_MIXMODE:
        {
	  UInt mm = ParseMixmode (reporter, params);
	  if (mm == (UInt)~0)
	  {
	    spr2dLook->DecRef ();
	    fact->DecRef ();
	    return NULL;
	  }
          spr2dLook->SetMixMode (mm);
	}
	break;
      case CS_TOKEN_UVANIMATION:
	{
	  ParseAnim (reporter, spr2dLook, name, params);
	}
        break;
    }
  }

  spr2dLook->DecRef ();
  return fact;
}

//---------------------------------------------------------------------------
csSprite2DFactorySaver::csSprite2DFactorySaver (iBase* pParent)
{
  SCF_CONSTRUCT_IBASE (pParent);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scfiComponent);
  reporter = NULL;
  plugin_mgr = NULL;
}

csSprite2DFactorySaver::~csSprite2DFactorySaver ()
{
  if (reporter) reporter->DecRef ();
  SCF_DEC_REF (plugin_mgr);
}

bool csSprite2DFactorySaver::Initialize (iObjectRegistry* object_reg)
{
  csSprite2DFactorySaver::object_reg = object_reg;
  plugin_mgr = CS_QUERY_REGISTRY (object_reg, iPluginManager);
  reporter = CS_QUERY_REGISTRY (object_reg, iReporter);
  return true;
}

#define MAXLINE 100 /* max number of chars per line... */

static void WriteMixmode(iStrVector *str, UInt mixmode)
{
  str->Push(csStrNew("  MIXMODE ("));
  if(mixmode&CS_FX_COPY) str->Push(csStrNew(" COPY ()"));
  if(mixmode&CS_FX_ADD) str->Push(csStrNew(" ADD ()"));
  if(mixmode&CS_FX_MULTIPLY) str->Push(csStrNew(" MULTIPLY ()"));
  if(mixmode&CS_FX_MULTIPLY2) str->Push(csStrNew(" MULTIPLY2 ()"));
  if(mixmode&CS_FX_KEYCOLOR) str->Push(csStrNew(" KEYCOLOR ()"));
  if(mixmode&CS_FX_TILING) str->Push(csStrNew(" TILING ()"));
  if(mixmode&CS_FX_TRANSPARENT) str->Push(csStrNew(" TRANSPARENT ()"));
  if(mixmode&CS_FX_ALPHA)
  {
    char buf[MAXLINE];
    sprintf(buf, "ALPHA (%g)", float(mixmode&CS_FX_MASK_ALPHA)/255.);
    str->Push(csStrNew(buf));
  }
  str->Push(csStrNew(")"));
}

void csSprite2DFactorySaver::WriteDown (iBase* obj, iStrVector * str,
  iEngine* /*engine*/)
{
  iSprite2DFactoryState *state =
    SCF_QUERY_INTERFACE (obj, iSprite2DFactoryState);
  char buf[MAXLINE];

  sprintf(buf, "MATERIAL (%s)\n", state->GetMaterialWrapper()->
    QueryObject ()->GetName());
  str->Push(csStrNew(buf));
  if(state->GetMixMode() != CS_FX_COPY)
  {
    WriteMixmode(str, state->GetMixMode());
  }
  sprintf(buf, "LIGHTING (%s)\n", state->HasLighting()?"true":"false");
  str->Push(csStrNew(buf));

  state->DecRef();
}
//---------------------------------------------------------------------------

csSprite2DLoader::csSprite2DLoader (iBase* pParent)
{
  SCF_CONSTRUCT_IBASE (pParent);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scfiComponent);
  reporter = NULL;
  plugin_mgr = NULL;
}

csSprite2DLoader::~csSprite2DLoader ()
{
  SCF_DEC_REF (plugin_mgr);
  if (reporter) reporter->DecRef ();
}

bool csSprite2DLoader::Initialize (iObjectRegistry* object_reg)
{
  csSprite2DLoader::object_reg = object_reg;
  plugin_mgr = CS_QUERY_REGISTRY (object_reg, iPluginManager);
  reporter = CS_QUERY_REGISTRY (object_reg, iReporter);
  return true;
}

iBase* csSprite2DLoader::Parse (const char* string, iEngine* engine,
	iBase* context)
{
  CS_TOKEN_TABLE_START (commands)
    CS_TOKEN_TABLE (FACTORY)
    CS_TOKEN_TABLE (VERTICES)
    CS_TOKEN_TABLE (UV)
    CS_TOKEN_TABLE (MATERIAL)
    CS_TOKEN_TABLE (MIXMODE)
    CS_TOKEN_TABLE (COLORS)
    CS_TOKEN_TABLE (LIGHTING)
    CS_TOKEN_TABLE (ANIMATE)
  CS_TOKEN_TABLE_END

  char* name;
  long cmd;
  char* params;
  char str[255];

  iMeshWrapper* imeshwrap = SCF_QUERY_INTERFACE (context, iMeshWrapper);
  CS_ASSERT (imeshwrap != NULL);
  imeshwrap->DecRef ();

  iMeshObject* mesh = NULL;
  iSprite2DState* spr2dLook = NULL;
  csColoredVertices* verts = NULL;

  char* buf = (char*)string;
  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    if (!params)
    {
      ReportError (reporter,
		"crystalspace.sprite2dloader.parse.badformat",
		"Bad format while parsing sprite2d!");
      if (spr2dLook) spr2dLook->DecRef ();
      return NULL;
    }
    switch (cmd)
    {
      case CS_TOKEN_FACTORY:
	{
          csScanStr (params, "%s", str);
	  iMeshFactoryWrapper* fact = engine->GetMeshFactories ()
	  	->FindByName (str);
	  if (!fact)
	  {
      	    ReportError (reporter,
		"crystalspace.sprite2dloader.parse.unknownfactory",
		"Couldn't find factory '%s'!", str);
	    if (spr2dLook) spr2dLook->DecRef ();
	    return NULL;
	  }
	  mesh = fact->GetMeshObjectFactory ()->NewInstance ();
	  imeshwrap->SetFactory (fact);
          spr2dLook = SCF_QUERY_INTERFACE (mesh, iSprite2DState);
	  verts = &(spr2dLook->GetVertices ());
	}
	break;
      case CS_TOKEN_MATERIAL:
	{
          csScanStr (params, "%s", str);
          iMaterialWrapper* mat = engine->GetMaterialList ()->
	  	FindByName (str);
	  if (!mat)
	  {
      	    ReportError (reporter,
		"crystalspace.sprite2dloader.parse.unknownmaterial",
		"Couldn't find material '%s'!", str);
            mesh->DecRef ();
	    if (spr2dLook) spr2dLook->DecRef ();
            return NULL;
	  }
	  spr2dLook->SetMaterialWrapper (mat);
	}
	break;
      case CS_TOKEN_MIXMODE:
        {
	  UInt mm = ParseMixmode (reporter, params);
	  if (mm == (UInt)~0)
	  {
	    if (spr2dLook) spr2dLook->DecRef ();
	    mesh->DecRef ();
	    return NULL;
	  }
          spr2dLook->SetMixMode (mm);
	}
	break;
      case CS_TOKEN_VERTICES:
        {
          float list[100];
	  int num;
          csScanStr (params, "%F", list, &num);
	  num /= 2;
	  verts->SetLength (num);
	  int i;
      for (i = 0 ; i < num ; i++)
	  {
	    (*verts)[i].pos.x = list[i*2+0];
	    (*verts)[i].pos.y = list[i*2+1];
	    (*verts)[i].color_init.Set (0, 0, 0);
	    (*verts)[i].color.Set (0, 0, 0);
	  }
        }
        break;
      case CS_TOKEN_UV:
        {
          float list[100];
	  int num;
          csScanStr (params, "%F", list, &num);
	  num /= 2;
	  verts->SetLength (num);
	  int i;
      for (i = 0 ; i < num ; i++)
	  {
	    (*verts)[i].u = list[i*2+0];
	    (*verts)[i].v = list[i*2+1];
	  }
        }
        break;
      case CS_TOKEN_LIGHTING:
        {
          bool do_lighting;
          csScanStr (params, "%b", &do_lighting);
          spr2dLook->SetLighting (do_lighting);
        }
        break;
      case CS_TOKEN_COLORS:
        {
          float list[100];
	  int num;
          csScanStr (params, "%F", list, &num);
	  num /= 3;
	  verts->SetLength (num);
	  int i;
      for (i = 0 ; i < num ; i++)
	  {
	    (*verts)[i].color_init.red = list[i*3+0];
	    (*verts)[i].color_init.green = list[i*3+1];
	    (*verts)[i].color_init.blue = list[i*3+2];
	  }
        }
        break;
      case CS_TOKEN_ANIMATE:
        {
          bool loop;
	  int type;
          csScanStr (params, "%s, %d, %b", str, &type, &loop);
	  iSprite2DUVAnimation *ani = spr2dLook->GetUVAnimation (str);
	  if (ani)
	    spr2dLook->SetUVAnimation (str, type, loop);
	  else
    	  {
	    ReportError (reporter,
		"crystalspace.sprite2dloader.parse.uvanim",
		"UVAnimation '%s' not found!", str);
	    if (spr2dLook) spr2dLook->DecRef ();
	    mesh->DecRef ();
	    return NULL;
	  }
        }
        break;
    }
  }

  if (spr2dLook) spr2dLook->DecRef ();
  return mesh;
}

//---------------------------------------------------------------------------

csSprite2DSaver::csSprite2DSaver (iBase* pParent)
{
  SCF_CONSTRUCT_IBASE (pParent);
  SCF_CONSTRUCT_EMBEDDED_IBASE(scfiComponent);
  reporter = NULL;
  plugin_mgr = NULL;
}

csSprite2DSaver::~csSprite2DSaver ()
{
  SCF_DEC_REF (plugin_mgr);
  if (reporter) reporter->DecRef ();
}

bool csSprite2DSaver::Initialize (iObjectRegistry* object_reg)
{
  csSprite2DSaver::object_reg = object_reg;
  plugin_mgr = CS_QUERY_REGISTRY (object_reg, iPluginManager);
  reporter = CS_QUERY_REGISTRY (object_reg, iReporter);
  return true;
}

void csSprite2DSaver::WriteDown (iBase* obj, iStrVector *str,
  iEngine* /*engine*/)
{
  iFactory *fact = SCF_QUERY_INTERFACE (this, iFactory);
  iSprite2DState *state = SCF_QUERY_INTERFACE (obj, iSprite2DState);
  char buf[MAXLINE];
  char name[MAXLINE];

  csFindReplace(name, fact->QueryDescription (), "Saver", "Loader", MAXLINE);
  sprintf(buf, "FACTORY ('%s')\n", name);
  str->Push(csStrNew(buf));

  sprintf(buf, "MATERIAL (%s)\n", state->GetMaterialWrapper()->
    QueryObject ()->GetName());
  str->Push(csStrNew(buf));
  sprintf(buf, "LIGHTING (%s)\n", state->HasLighting()?"true":"false");
  str->Push(csStrNew(buf));
  if(state->GetMixMode() != CS_FX_COPY)
  {
    WriteMixmode(str, state->GetMixMode());
  }

  csColoredVertices& vs = state->GetVertices();
  int i;
  str->Push(csStrNew("VERTICES("));
  for(i=0; vs.Length(); i++)
  {
    sprintf(buf, "%g,%g%s", vs[i].pos.x, vs[i].pos.y,
      (i==vs.Length()-1)?"":", ");
    str->Push(csStrNew(buf));
  }
  str->Push(csStrNew(")\n"));

  str->Push(csStrNew("UV("));
  for(i=0; vs.Length(); i++)
  {
    sprintf(buf, "%g,%g%s", vs[i].u, vs[i].v, (i==vs.Length()-1)?"":", ");
    str->Push(csStrNew(buf));
  }
  str->Push(csStrNew(")\n"));

  str->Push(csStrNew("COLORS("));
  for(i=0; vs.Length(); i++)
  {
    sprintf(buf, "%g,%g,%g%s", vs[i].color_init.red, vs[i].color_init.green,
      vs[i].color_init.blue, (i==vs.Length()-1)?"":", ");
    str->Push(csStrNew(buf));
  }
  str->Push(csStrNew(")\n"));

  fact->DecRef();
  state->DecRef();
}
